from collections import defaultdict
from itertools import product
from typing import TYPE_CHECKING

import pandas as pd
from django.db.models import Exists, F, OuterRef, Q
from django.utils import timezone
from rest_framework import status
from rest_framework.decorators import action
from rest_framework.response import Response
from wbcore import viewsets
from wbcore.contrib.notifications.dispatch import send_notification
from wbcore.utils.strings import format_number

from wbportfolio.models import AccountReconciliation, AccountReconciliationLine
from wbportfolio.models.reconciliations.account_reconciliation_lines import (
    AccountReconciliationLineQuerySet,
)
from wbportfolio.serializers import (
    AccountReconciliationLineModelSerializer,
    AccountReconciliationModelSerializer,
)
from wbportfolio.viewsets.configs.buttons import (
    AccountReconciliationButtonViewConfig,
    AccountReconciliationLineButtonViewConfig,
)
from wbportfolio.viewsets.configs.display import (
    AccountReconciliationDisplayViewConfig,
    AccountReconciliationLineDisplayViewConfig,
)
from wbportfolio.viewsets.configs.endpoints.reconciliations import (
    AccountReconciliationEndpointViewConfig,
    AccountReconciliationLineEndpointViewConfig,
)

if TYPE_CHECKING:
    from rest_framework.request import Request


class AccountReconciliationModelViewSet(viewsets.ModelViewSet):
    INSTANCE_DOCUMENTATION = "wbportfolio/markdown/documentation/account_holding_reconciliation.md"

    queryset = AccountReconciliation.objects.all()
    serializer_class = AccountReconciliationModelSerializer
    display_config_class = AccountReconciliationDisplayViewConfig
    button_config_class = AccountReconciliationButtonViewConfig
    endpoint_config_class = AccountReconciliationEndpointViewConfig

    filterset_fields = {
        "account": ["exact"],
        "creator": ["exact"],
        "approved_by": ["exact"],
        "reconciliation_date": ["gte", "exact", "lte"],
        "approved_dt": ["gte", "lte"],
    }
    ordering = ordering_fields = ["-reconciliation_date"]

    def get_queryset(self):
        return (
            AccountReconciliation.objects.filter_for_user(self.request.user)
            .annotate(
                unequal_exists=Exists(
                    AccountReconciliationLine.objects.filter(
                        Q(reconciliation_id=OuterRef("pk")) & ~Q(shares=F("shares_external"))
                    )
                )
            )
            .select_related("account")
            .select_related("creator")
            .select_related("creator__profile")
            .select_related("approved_by")
            .select_related("approved_by__profile")
        )

    @action(methods=["PATCH"], detail=True)
    def agree_customer(self, request: "Request", pk: int | None = None) -> Response:
        reconciliation = self.get_object()
        if reconciliation and isinstance(reconciliation, AccountReconciliation):
            reconciliation.approved_by = request.user
            reconciliation.approved_dt = timezone.now()
            reconciliation.save()
            send_notification(
                code="wbportfolio.account_reconciliation.notify",
                title=f"{reconciliation.approved_by} has agreed to the calculations for the account {reconciliation.account}.",
                body=f"The account {reconciliation.account} has been reconciled and the calculations have been agreed to by {reconciliation.approved_by}. You may proceed.",
                user=reconciliation.creator,
                reverse_name="wbportfolio:accountreconciliation-detail",
                reverse_args=[reconciliation.id],
            )
            return Response({"__notification": "You have aggreed to the calculations, thank you."})
        return Response(
            {"__notification": "There has been an issue with agreeing to the calculations."},
            status=status.HTTP_400_BAD_REQUEST,
        )

    @action(methods=["PATCH"], detail=True)
    def disagree_customer(self, request: "Request", pk: int | None = None) -> Response:
        reconciliation = self.get_object()
        if reconciliation and isinstance(reconciliation, AccountReconciliation):
            send_notification(
                code="wbportfolio.account_reconciliation.notify",
                title=f"{request.user} has disagreed with your calculations for the account {reconciliation.account}",
                body=f"Please reach out to {request.user} to adjust the reconciliation.",
                user=reconciliation.creator,
                reverse_name="wbportfolio:accountreconciliation-detail",
                reverse_args=[reconciliation.id],
            )
        return Response()

    @action(methods=["PATCH"], detail=True)
    def recompute(self, request: "Request", pk: int | None = None) -> Response:
        AccountReconciliationLine.objects.update_or_create_for_reconciliation(self.get_object())
        return Response()

    @action(methods=["PATCH"], detail=True)
    def notify(self, request: "Request", pk: int | None = None) -> Response:
        try:
            self.get_object().notify(update=True, silent_errors=False)
            return Response()
        except ValueError as e:
            return Response(str(e), status=status.HTTP_400_BAD_REQUEST)


class AccountReconciliationLineModelViewSet(viewsets.ModelViewSet):
    DEPENDANT_IDENTIFIER = ["wbportfolio:accountreconciliation"]

    queryset = AccountReconciliationLine.objects.all()
    serializer_class = AccountReconciliationLineModelSerializer
    display_config_class = AccountReconciliationLineDisplayViewConfig
    button_config_class = AccountReconciliationLineButtonViewConfig
    endpoint_config_class = AccountReconciliationLineEndpointViewConfig

    def get_aggregates(self, queryset, paginated_queryset):
        fields = [
            "assets_under_management",
            "assets_under_management_external",
            "assets_under_management_diff",
        ]
        df = pd.DataFrame(queryset.values("currency_key", *fields))
        aggregates = defaultdict(dict)
        if not df.empty:
            for field in fields:
                aggregates[field]["Your"] = "Total AuM"

            for currency_symbol, field in product(df.currency_key.unique(), fields):
                dff = df.loc[df["currency_key"] == currency_symbol]
                aggregates[field][f"{currency_symbol}"] = format_number(dff[field].sum())

        return aggregates

    def get_queryset(self) -> "AccountReconciliationLineQuerySet":
        queryset: "AccountReconciliationLineQuerySet" = super().get_queryset()

        if accountreconciliation_id := self.kwargs.get("accountreconciliation_id"):
            queryset = queryset.filter(reconciliation_id=accountreconciliation_id)

        return (
            queryset.annotate_currency()
            .annotate_currency_key()
            .annotate_assets_under_management()
            .annotate_assets_under_management_external()
            .annotate_shares_diff()
            .annotate_nominal_value_diff()
            .annotate_assets_under_management_diff()
            .annotate_pct_diff()
            .annotate_is_equal()
            .select_related("reconciliation")
            .select_related("reconciliation__account")
            .select_related("product")
        )



import json
from datetime import datetime
from io import BytesIO
from typing import Optional

from django.db import models
from pandas.tseries.offsets import BDay
from wbcore.contrib.io.backends import AbstractDataBackend, register

from wbportfolio.api_clients.ubs import UBSNeoAPIClient

from .mixin import DataBackendMixin


@register("Instrument Prices", provider_key="ubs", save_data_in_import_source=True, passive_only=True)
class DataBackend(DataBackendMixin, AbstractDataBackend):
    def __init__(
        self, import_credential: Optional[models.Model] = None, ubs_bank: Optional[models.Model] = None, **kwargs
    ):
        if not ubs_bank:
            raise ValueError("The ubs company objects needs to be passed to this backend")
        self.ubs_bank = ubs_bank
        if not import_credential or not import_credential.authentication_token:
            raise ValueError("UBS backend needs a valid import credential object")
        self.authentication_token = import_credential.authentication_token.replace("Bearer ", "")
        self.token_expiry_date = import_credential.validity_end

    def get_files(
        self,
        execution_time: datetime,
        obj_external_ids: list[str] = None,
        **kwargs,
    ) -> BytesIO:
        execution_date = (execution_time - BDay(1)).date()
        if obj_external_ids:
            client = UBSNeoAPIClient(self.authentication_token, self.token_expiry_date)
            for external_id in obj_external_ids:
                res_json = client.validate_response(client.get_portfolio_at_date(external_id, execution_date))
                res_json.pop("constituents", None)
                if res_json:
                    content_file = BytesIO()
                    content_file.write(json.dumps(res_json).encode())
                    file_name = f"ubs_instrument_price_{external_id}_{execution_date:%Y-%m-%d}_{int(datetime.timestamp(execution_time))}.json"
                    yield file_name, content_file



import datetime
import logging
import operator
from decimal import Decimal
from functools import reduce
from typing import Optional

import numpy as np
from celery import chain, shared_task
from django.db import models
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver
from django_fsm import FSMField, transition
from pandas.tseries.offsets import BDay
from wbcore.contrib.authentication.models import User
from wbcore.contrib.icons import WBIcon
from wbcore.contrib.io.mixins import ImportMixin
from wbcore.contrib.notifications.dispatch import send_notification
from wbcore.enums import RequestType
from wbcore.metadata.configs.buttons import ActionButton
from wbcore.signals import pre_merge
from wbcore.workers import Queue
from wbfdm.models.instruments.instruments import Instrument, import_prices_as_task

from wbportfolio.import_export.handlers.adjustment import AdjustmentImportHandler
from wbportfolio.models.roles import PortfolioRole

logger = logging.getLogger("pms")


class Adjustment(ImportMixin, models.Model):
    import_export_handler_class = AdjustmentImportHandler

    class Status(models.TextChoices):
        PENDING = "PENDING", "Pending"
        APPLIED = "APPLIED", "Applied"
        DENIED = "DENIED", "Denied"

    status = FSMField(default=Status.PENDING, choices=Status.choices, verbose_name="Status")

    date = models.DateField(verbose_name="Executive Date")
    instrument = models.ForeignKey(
        "wbfdm.Instrument",
        on_delete=models.CASCADE,
        related_name="pms_adjustments",
        limit_choices_to=models.Q(("children__isnull", True)),
    )
    factor = models.DecimalField(default=Decimal(1), max_digits=30, decimal_places=17, verbose_name="Factor")
    cumulative_factor = models.DecimalField(
        default=Decimal(1), max_digits=30, decimal_places=17, verbose_name="Cumulative Factor"
    )
    last_handler = models.ForeignKey(
        "directory.Person", blank=True, null=True, on_delete=models.SET_NULL, related_name="last_handled_adjustments"
    )

    def __str__(self) -> str:
        return f"Instrument {str(self.instrument)} (date: {self.date}, factor: {self.factor})"

    @property
    def adjustment_date(self) -> datetime.date:
        return (self.date - BDay(1)).date()

    def save(self, *args, **kwargs):
        if not self.cumulative_factor or self.cumulative_factor == Decimal(1):
            self.cumulative_factor = self.get_cumulative_factor()
        super().save(*args, **kwargs)

    def get_cumulative_factor(self) -> Decimal:
        applied_parent_adjustments_factors = list(
            Adjustment.objects.exclude(id=self.id)
            .filter(status=self.Status.APPLIED, instrument=self.instrument, date__gte=self.date)
            .values_list("factor", flat=True)
        )
        return Decimal(np.prod([*applied_parent_adjustments_factors]))

    def automatically_applied_adjustments_on_assets(self) -> bool:
        return reduce(operator.or_, [1 / self.factor % i == 0 for i in range(2, 5)])

    def apply_adjustment_on_assets(self):
        self.instrument.assets.filter(
            models.Q(date__lte=self.adjustment_date)
            | (models.Q(applied_adjustment__isnull=True) & models.Q(applied_adjustment__date__lt=self.date))
        ).update(applied_adjustment=self)

    def revert_adjustment_on_assets(self):
        self.adjustmented_assets.update(applied_adjustment=None)
        future_adjustments = self.instrument.pms_adjustments.filter(date__gt=self.date)
        if future_adjustments.exists():
            future_adjustments.earliest(
                "date"
            ).apply_adjustment_on_assets()  # We reevalute the next future adjustment in order to check if this one can replace the reverted adjustment
        self.status = self.Status.PENDING
        self.save()

    @transition(
        field=status,
        source=[Status.PENDING],
        target=Status.APPLIED,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(user.profile),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:adjustment",),
                icon=WBIcon.APPROVE.icon,
                key="apply",
                label="Apply",
                action_label="Apply",
                # description_fields="<p>Start: {{start}}</p><p>End: {{end}}</p><p>Title: {{title}}</p>",
            )
        },
    )
    def apply(self, by: Optional["User"] = None, description: Optional[str] = None, **kwargs):
        if profile := getattr(by, "profile", None):
            self.last_handler = profile
        apply_adjustment_on_assets_as_task.delay(self.id)

    @transition(
        field=status,
        source=[Status.PENDING],
        target=Status.DENIED,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(user.profile),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:adjustment",),
                icon=WBIcon.DENY.icon,
                key="deny",
                label="Deny",
                action_label="Deny",
                # description_fields="<p>Start: {{start}}</p><p>End: {{end}}</p><p>Title: {{title}}</p>",
            )
        },
    )
    def deny(self, by=None, description=None, **kwargs):
        if profile := getattr(by, "profile", None):
            self.last_handler = profile

    @transition(
        field=status,
        source=[Status.APPLIED],
        target=Status.PENDING,
        permission=lambda instance, user: PortfolioRole.is_portfolio_manager(user.profile),
        custom={
            "_transition_button": ActionButton(
                method=RequestType.PATCH,
                identifiers=("wbportfolio:adjustment",),
                icon=WBIcon.UNDO.icon,
                key="revert",
                label="Revert",
                action_label="Revert",
            )
        },
    )
    def revert(self, by=None, description=None, **kwargs):
        if profile := getattr(by, "profile", None):
            self.last_handler = profile
        revert_adjustment_on_assets_as_task.delay(self.id)

    def import_instrument_prices(self):
        # If the adjusted instrument is not within any portfolio, we automatically approve (and applied) the adjustments
        # We reimport prices when a adjustment comes in, not really the most efficient but at least, we are sure we keep having corrected price
        chain(
            import_prices_as_task.si(self.instrument.id, clear=True),
            post_adjustment_on_prices.si(self.id, automatically_confirm_approve_adjustment_on_assets=True),
        ).apply_async()

    class Meta:
        verbose_name = "Datum Adjustment"
        verbose_name_plural = "Data Adjustment"
        constraints = [
            models.UniqueConstraint(
                fields=["date", "instrument"],
                name="unique_date_instrument",
            ),
        ]
        indexes = [
            models.Index(
                name="instrument_adjustment_idx",
                fields=["instrument", "-date"],
            )
        ]

    @classmethod
    def get_endpoint_basename(cls):
        return "wbportfolio:adjustment"


# Shared tasks


@shared_task(queue=Queue.BACKGROUND.value)
def apply_adjustment_on_assets_as_task(adjustment_id):
    adjustment = Adjustment.objects.get(id=adjustment_id)
    adjustment.apply_adjustment_on_assets()


@shared_task(queue=Queue.BACKGROUND.value)
def revert_adjustment_on_assets_as_task(adjustment_id):
    adjustment = Adjustment.objects.get(id=adjustment_id)
    adjustment.revert_adjustment_on_assets()


@shared_task(queue=Queue.BACKGROUND.value)
def post_adjustment_on_prices(adjustment_id, automatically_confirm_approve_adjustment_on_assets: bool | None = False):
    # We just notified the concerned users that a adjustment has been processed, prices have been reimported and as there
    # are assets on the instrument, manually confirmation is needed before apply the adjustment
    adjustment = Adjustment.objects.get(id=adjustment_id)
    if adjustment.instrument.assets.filter(date__lte=adjustment.date).exists():
        if (
            adjustment.automatically_applied_adjustments_on_assets()
            or automatically_confirm_approve_adjustment_on_assets
        ):
            adjustment.apply_adjustment_on_assets()
            adjustment.status = Adjustment.Status.APPLIED
        else:
            for user in User.objects.filter(profile__in=PortfolioRole.portfolio_managers(), is_active=True):
                send_notification(
                    code="wbportfolio.adjustment.add",
                    title="A new adjustment was imported",
                    body=f"A adjustment for {str(adjustment.instrument)} with ex Date  {adjustment.date} and factor {adjustment.factor} was imported. Please confirm (to apply) or deny it.",
                    user=user,
                    reverse_name="wbportfolio:adjustment-detail",
                    reverse_args=[adjustment.id],
                )
    else:
        adjustment.status = Adjustment.Status.APPLIED
    adjustment.save()


# Receivers


@receiver(post_delete, sender=Adjustment)
def post_delete_adjustment(sender, instance, **kwargs):
    for adjustment in Adjustment.objects.filter(instrument=instance.instrument).order_by("date"):
        adjustment.save()


@receiver(post_save, sender="wbportfolio.Adjustment")
def post_save_adjustment(sender, instance, created, raw, **kwargs):
    if created and (instance.instrument.assets.exists() or instance.instrument.prices.exists()):
        instance.import_instrument_prices()


@receiver(pre_merge, sender="wbfdm.Instrument")
def pre_merge_instrument(sender: models.Model, merged_object: "Instrument", main_object: "Instrument", **kwargs):
    """
    Simply reassign the adjustments of the merged instrument to the main instrument if they don't already exist for that day, otherwise simply delete them
    """
    merged_object.pms_adjustments.annotate(
        already_exists=models.Exists(Adjustment.objects.filter(date=models.OuterRef("date"), instrument=main_object))
    ).filter(already_exists=True).delete()
    merged_object.pms_adjustments.update(instrument=main_object)



import logging
from datetime import date
from decimal import Decimal
from typing import Iterator

from celery import shared_task
from django.db.models import F, QuerySet, Window
from django.db.models.functions import RowNumber
from tqdm import tqdm
from wbcore.workers import Queue
from wbfdm.models import Instrument

from wbportfolio.models import AssetPosition, Index, Order, OrderProposal, Portfolio, Product

logger = logging.getLogger("pms")


def get_casted_portfolio_instrument(instrument: Instrument) -> Product | Index | None:
    try:
        return Product.objects.get(id=instrument.id)
    except Product.DoesNotExist:
        try:
            return Index.objects.get(id=instrument.id)
        except Index.DoesNotExist:
            return None


def get_adjusted_shares(old_shares: Decimal, old_price: Decimal, new_price: Decimal) -> Decimal:
    return old_shares * (old_price / new_price)


def adjust_assets(qs: Iterator[AssetPosition], underlying_quote: Instrument):
    objs = []
    logger.info("adjusting asset positions...")
    for a in qs:
        old_price: Decimal = a.initial_price
        a.initial_price = a.underlying_instrument = a.underlying_quote_price = None
        a.underlying_quote = underlying_quote
        a.pre_save()
        if a.initial_shares and a.initial_price and old_price != a.initial_price:
            a.initial_shares = get_adjusted_shares(a.initial_shares, old_price, a.initial_price)
        objs.append(a)
    AssetPosition.objects.bulk_update(
        objs,
        ["underlying_quote", "underlying_quote_price", "underlying_instrument", "initial_price", "initial_shares"],
        batch_size=1000,
    )


def adjust_orders(qs: Iterator[Order], underlying_quote: Instrument):
    objs = []
    logger.info("adjusting orders...")
    for o in qs:
        old_price: Decimal = o.price
        o.underlying_instrument = underlying_quote
        o.set_price()
        if o.price and old_price != o.price and o.shares:
            o.shares = get_adjusted_shares(o.shares, old_price, o.price)
        objs.append(o)
    Order.objects.bulk_update(objs, ["price", "shares", "underlying_instrument"], batch_size=1000)


def adjust_quote(
    old_quote: Instrument,
    new_quote: Instrument,
    adjust_after: date | None = None,
    only_portfolios: QuerySet[Portfolio] | None = None,
    debug: bool = False,
):
    if old_quote.currency != new_quote.currency:
        raise ValueError("cannot safely switch quotes that are not of the same currency")
    assets_to_change = AssetPosition.objects.filter(underlying_quote=old_quote)
    orders_to_change = Order.objects.filter(underlying_instrument=old_quote)
    new_quote.import_prices()
    if adjust_after:
        assets_to_change = assets_to_change.filter(date__gt=adjust_after)
        orders_to_change = orders_to_change.filter(value_date__gt=adjust_after)
    if only_portfolios is not None:
        assets_to_change = assets_to_change.filter(portfolio__in=only_portfolios)
        orders_to_change = orders_to_change.filter(order_proposal__portfolio__in=only_portfolios)
    if debug:
        assets_to_change = tqdm(assets_to_change, total=assets_to_change.count())
        orders_to_change = tqdm(orders_to_change, total=orders_to_change.count())

    # gather the list of order proposal to replay (if the quote led to missing position, we want to replay it to correct automatically the issue)
    latest_orders = orders_to_change.annotate(
        row_number=Window(
            expression=RowNumber(), partition_by=[F("order_proposal__portfolio")], order_by=F("value_date").desc()
        )
    ).filter(row_number=1)
    order_proposals_to_replay = OrderProposal.objects.filter(
        portfolio__is_manageable=True, id__in=latest_orders.values("order_proposal")
    )

    # Adjust assets to the new quote
    adjust_assets(assets_to_change, new_quote)

    # Adjust orders to the new quote
    adjust_orders(orders_to_change, new_quote)

    # replay latest order proposal
    for op in order_proposals_to_replay:
        op.replay(reapply_order_proposal=True)


@shared_task(queue=Queue.BACKGROUND.value)
def adjust_quote_as_task(
    old_quote_id: int, new_quote_id: int, adjust_after: date | None = None, only_portfolio_ids: list[int] | None = None
):
    old_quote = Instrument.objects.get(id=old_quote_id)
    new_quote = Instrument.objects.get(id=new_quote_id)
    only_portfolios = Portfolio.objects.filter(id__in=only_portfolio_ids) if only_portfolio_ids else None
    adjust_quote(old_quote, new_quote, adjust_after=adjust_after, only_portfolios=only_portfolios)



from django.contrib import admin

from wbportfolio.models import (
    DailyPortfolioCashFlow,
    InstrumentPortfolioThroughModel,
    Portfolio,
    PortfolioBankAccountThroughModel,
    PortfolioCashTarget,
    PortfolioPortfolioThroughModel,
    PortfolioSwingPricing,
)


@admin.register(PortfolioBankAccountThroughModel)
class PortfolioBankAccountThroughModelModelAdmin(admin.ModelAdmin):
    autocomplete_fields = ["bank_account", "portfolio"]
    list_display = ["bank_account", "portfolio", "portfolio_bank_account_type"]


class InstrumentPortfolioThroughModelAdmin(admin.TabularInline):
    model = InstrumentPortfolioThroughModel
    fk_name = "portfolio"
    raw_id_fields = ["portfolio", "instrument"]


class PortfolioPortfolioThroughModelInlineAdmin(admin.TabularInline):
    model = PortfolioPortfolioThroughModel
    fk_name = "portfolio"
    fields = ["dependency_portfolio", "type"]
    autocomplete_fields = ["portfolio", "dependency_portfolio"]


@admin.register(PortfolioPortfolioThroughModel)
class PortfolioPortfolioThroughModelModelAdmin(admin.ModelAdmin):
    list_display = ["portfolio", "dependency_portfolio", "type"]


class PortfolioBankAccountThroughModelAdmin(admin.TabularInline):
    model = PortfolioBankAccountThroughModel
    fields = ["bank_account", "portfolio_bank_account_type"]


@admin.register(Portfolio)
class PortfolioModelAdmin(admin.ModelAdmin):
    search_fields = ("name",)

    fieldsets = (
        (
            "Main Information",
            {
                "fields": (
                    ("name", "updated_at"),
                    ("currency", "hedged_currency"),
                    (
                        "invested_timespan",
                        "is_manageable",
                        "is_active",
                    ),
                    ("is_tracked", "only_keep_essential_positions", "only_weighting", "is_lookthrough"),
                )
            },
        ),
        (
            "OMS",
            {
                "fields": (
                    "default_order_proposal_min_order_value",
                    "default_order_proposal_min_weighting",
                    "default_order_proposal_total_cash_weight",
                )
            },
        ),
    )
    readonly_fields = ["updated_at"]
    list_display = (
        "is_active",
        "name",
        "currency",
        "hedged_currency",
        "is_manageable",
        "is_tracked",
        "only_keep_essential_positions",
        "is_lookthrough",
    )
    inlines = [
        PortfolioBankAccountThroughModelAdmin,
        InstrumentPortfolioThroughModelAdmin,
        # PortfolioInstrumentPreferredClassificationThroughInlineModelAdmin,
        PortfolioPortfolioThroughModelInlineAdmin,
    ]
    raw_id_fields = [
        "depends_on",
        "preferred_instrument_classifications",
        "currency",
        "hedged_currency",
    ]
    autocomplete_fields = ["currency", "hedged_currency"]

    def get_queryset(self, request):
        return Portfolio.all_objects.all()

    def update_preferred_classification_per_instrument(self, request, queryset):
        for portfolio in queryset.all():
            portfolio.update_preferred_classification_per_instrument()

    actions = [update_preferred_classification_per_instrument]


@admin.register(DailyPortfolioCashFlow)
class DailyPortfolioCashFlowModelAdmin(admin.ModelAdmin):
    autocomplete_fields = ["portfolio"]
    raw_id_fields = ["import_source"]
    list_display = [
        "portfolio",
        "value_date",
    ]
    readonly_fields = [
        "estimated_total_assets",
        "true_cash",
        "true_cash_pct",
        "cash_pct",
        "target_cash",
        "excess_cash",
        "min_target_cash_pct",
        "target_cash_pct",
        "max_target_cash_pct",
        "proposed_rebalancing",
        "rebalancing",
    ]


@admin.register(PortfolioCashTarget)
class PortfolioCashTargetModelAdmin(admin.ModelAdmin):
    autocomplete_fields = ["portfolio"]


@admin.register(PortfolioSwingPricing)
class PortfolioSwingPricingModelAdmin(admin.ModelAdmin):
    autocomplete_fields = ["portfolio"]



from django.db.models import Q
from dynamic_preferences.registries import global_preferences_registry
from wbfdm.models import Instrument


def get_timedelta_import_instrument_price():
    return global_preferences_registry.manager()["wbportfolio__timedelta_import_instrument_price"]


def filter_active_instruments(_date, queryset=None):
    if not queryset:
        queryset = Instrument.objects
    queryset = queryset.filter(Q(delisted_date__isnull=True) | Q(delisted_date__gte=_date))
    queryset = queryset.filter(
        Q(refinitiv_mnemonic_code__isnull=False) | Q(refinitiv_identifier_code__isnull=False) | Q(isin__isnull=False)
    )
    return queryset.distinct()


def chunked_queryset(queryset, chunk_size):
    """Slice a queryset into chunks."""

    start_pk = 0
    queryset = queryset.order_by("pk")

    while True:
        # No entry left
        if not queryset.filter(pk__gt=start_pk).exists():
            break

        try:
            # Fetch chunk_size entries if possible
            end_pk = queryset.filter(pk__gt=start_pk).values_list("pk", flat=True)[chunk_size - 1]

            # Fetch rest entries if less than chunk_size left
        except IndexError:
            end_pk = queryset.values_list("pk", flat=True).last()

        yield queryset.filter(pk__gt=start_pk).filter(pk__lte=end_pk)

        start_pk = end_pk



# Generated by Django 5.0.14 on 2025-08-18 08:34

from decimal import Decimal
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('wbportfolio', '0087_product_order_routing_custodian_adapter'),
    ]

    operations = [
        migrations.AddField(
            model_name='orderproposal',
            name='total_effective_portfolio_contribution',
            field=models.DecimalField(decimal_places=16, default=Decimal('1'), max_digits=19),
        ),
    ]
